import asyncio
import logging
import time

import orjson

from aiograpi import reqwests
from aiograpi.exceptions import (
    AboutUsError,
    AccountSuspended,
    ChallengeRequired,
    ClientBadRequestError,
    ClientConnectionError,
    ClientError,
    ClientForbiddenError,
    ClientJSONDecodeError,
    ClientLoginRequired,
    ClientNotFoundError,
    ClientThrottledError,
    ClientUnauthorizedError,
    TermsAccept,
    TermsUnblock,
)
from aiograpi.utils import random_delay

GRAPHQL_API_URL = "https://www.instagram.com/api/graphql"

GQL_STUFF = {
    "av": "17841464591314721",
    "__d": "www",
    "__user": "0",
    "__a": "1",
    "__req": "q",
    "__hs": "19768.HYP:instagram_web_pkg.2.1..0.1",
    "dpr": "2",
    "__ccg": "UNKNOWN",
    "__rev": "1011444902",
    "__s": "x82a1q:agr3gd:4nh4nl",
    "__hsi": "7335888108907652597",
    "__dyn": (
        "7xeUjG1mxu1syUbFp40NonwgU7SbzEdF8aUco2qwJxS0k24o0B-"
        "q1ew65xO0FE2awt81s8hwGwQwoEcE7O2l0Fwqo31w9O7U2cxe0E"
        "UjwGzEaE7622362W2K0zK5o4q3y1Sx-0iS2Sq2-azqwt8dUaob8"
        "2cwMwrUdUbGwmk0KU6O1FwlE6PhA6bxy4VUKUnAwHw"
    ),
    "__csr": (
        "g9cj5kxfs8lifTitQDqhdhalmDEAJaKBRJFdkAGHBkPy9HgCA-A"
        "rtucm5bCBBGpyAoz-mLJpXJufKWGQ9hHhAhnKECuFUZ3Q8Jkmmp"
        "eWyGAzkEj_CjyoZUgK-E8bwYzaxy00ktMGx20XU3gw4KAo3MChU"
        "jw3N80poolwiA1d7G2yu2ucxi1nwEw16OE1JsS043Etw63wkSEgg1Mu00yiU"
    ),
    "__comet_req": "7",
    "lsd": "6b2800R9u4biJOYjcdXFEI",
    "__spin_r": "1011444902",
    "__spin_b": "trunk",
    "__spin_t": "1708019550",
    "fb_api_caller_class": "RelayModern",
    "fb_api_req_friendly_name": "PolarisPostCommentsPaginationQuery",
    "server_timestamps": "true",
}


class GraphQLRequestMixin:
    _fb_dtsg = None
    graphql_requests_count = 0
    last_graphql_response = None
    last_graphql_json = {}
    request_logger = logging.getLogger("graphql_request")

    def __init__(self, *args, **kwargs):
        self.graphql = reqwests.Session()
        self.graphql.verify = False  # fix SSLError/HTTPSConnectionPool
        self.graphql.headers.update(
            {
                "Connection": "Keep-Alive",
                "Accept": "*/*",
                "Accept-Encoding": "gzip,deflate",
                "Accept-Language": "en-US,en;q=0.9",
                "origin": "https://www.instagram.com",
                "authority": "www.instagram.com",
                "sec-fetch-site": "same-origin",
                "User-Agent": (
                    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
                    "AppleWebKit/537.36 (KHTML, like Gecko) "
                    "Chrome/121.0.0.0 Safari/537.36"
                ),
            }
        )
        super().__init__(*args, **kwargs)

    @property
    async def fb_dtsg(self):
        if not self._fb_dtsg:
            self._fb_dtsg = await self.fetch_fb_dtsg()
        return self._fb_dtsg

    async def fetch_fb_dtsg(self):
        self.inject_sessionid_to_public()
        response = await self.graphql.get(GRAPHQL_API_URL)
        if html := response.content.decode():
            s = html.find("__eqmc")
            e = s + 1000
            eqmc = html[s:e]
            s, e = eqmc.find("{"), eqmc.find("</script>")
            eqmc = eqmc[s:e]
            eqmc = orjson.loads(eqmc)
            return eqmc["f"]
        return None

    async def graphql_request(
        self,
        data=None,
        params=None,
        headers=None,
        return_json=True,
        retries_count=1,
        retries_timeout=2,
    ):
        kwargs = dict(
            data=data,
            params=params,
            headers=headers,
            return_json=return_json,
        )
        if retries_count > 10:
            raise Exception("Retries count is too high")
        if retries_timeout > 600:
            raise Exception("Retries timeout is too high")
        self.inject_sessionid_to_public()
        for iteration in range(retries_count):
            try:
                if self.delay_range:
                    await random_delay(delay_range=self.delay_range)
                return await self._send_graphql_request(**kwargs)
            except (
                ClientLoginRequired,
                ClientNotFoundError,
                ClientBadRequestError,
            ) as e:
                raise e  # Stop retries
            except ClientError as e:
                msg = str(e)
                if all(
                    (
                        isinstance(e, ClientConnectionError),
                        "SOCKSHTTPSConnectionPool" in msg,
                        "Max retries exceeded with url" in msg,
                        "Failed to establish a new connection" in msg,
                    )
                ):
                    raise e
                if retries_count > iteration + 1:
                    await asyncio.sleep(retries_timeout)
                else:
                    raise e
                continue

    async def _send_graphql_request(
        self, data=None, params=None, headers=None, return_json=False
    ):
        self.last_graphql_response = None
        self.graphql_requests_count += 1
        if headers:
            self.graphql.headers.update(headers)
        if self.last_response_ts and (time.time() - self.last_response_ts) < 1.0:
            await asyncio.sleep(1.0)
        try:
            if data is not None:
                response = await self.graphql.post(
                    GRAPHQL_API_URL, data=data, params=params
                )
            else:
                response = await self.graphql.get(GRAPHQL_API_URL, params=params)
            self.request_logger.debug(
                "graphql_request %s: %s", response.status_code, response.url
            )
            self.request_logger.info(
                "GraphQL: [%s] [%s] %s %s",
                self.graphql.proxies.get("https"),
                response.status_code,
                "POST" if data else "GET",
                response.url,
            )
            self.last_graphql_response = response
            response.raise_for_status()
            if return_json:
                self.last_graphql_json = response.json()
                return self.last_graphql_json
            return response.text

        except orjson.JSONDecodeError as e:
            url = str(response.url)
            if "/login/" in url:
                raise ClientLoginRequired(e, response=response)
            elif "/challenge/" in url:
                raise ChallengeRequired(e, response=response)
            elif "/suspended/" in url:
                raise AccountSuspended(e, response=response)
            elif "/terms/unblock" in url:
                raise TermsUnblock(e, response=response)
            elif "/terms/accept" in url:
                raise TermsAccept(e, response=response)
            elif "/about-us" in url:
                raise AboutUsError(e, response=response)

            self.request_logger.error(
                "Status %s: JSONDecodeError in graphql_request (url=%s) >>> %s",
                response.status_code,
                url,
                response.text,
            )
            raise ClientJSONDecodeError(
                "JSONDecodeError {0!s} while opening {1!s}".format(e, url),
                response=response,
            )
        except reqwests.HTTPError as e:
            match getattr(self.last_graphql_response, "status_code", None):
                case 401:
                    exc = ClientUnauthorizedError
                case 403:
                    exc = ClientForbiddenError
                case 400:
                    exc = ClientBadRequestError
                case 429:
                    exc = ClientThrottledError
                case 404:
                    exc = ClientNotFoundError
                case _:
                    exc = ClientError
            raise exc(e, response=self.last_graphql_response)
        except (reqwests.ConnectError, reqwests.ReadError) as e:
            raise ClientConnectionError("{} {}".format(e.__class__.__name__, str(e)))
        finally:
            self.last_response_ts = time.time()



import logging
from urllib.parse import urlparse

from aiograpi.mixins.account import AccountMixin
from aiograpi.mixins.album import DownloadAlbumMixin, UploadAlbumMixin
from aiograpi.mixins.auth import LoginMixin
from aiograpi.mixins.bloks import BloksMixin
from aiograpi.mixins.challenge import ChallengeResolveMixin
from aiograpi.mixins.clip import DownloadClipMixin, UploadClipMixin
from aiograpi.mixins.collection import CollectionMixin
from aiograpi.mixins.comment import CommentMixin
from aiograpi.mixins.direct import DirectMixin
from aiograpi.mixins.fbsearch import FbSearchMixin
from aiograpi.mixins.graphql import GraphQLRequestMixin
from aiograpi.mixins.hashtag import HashtagMixin
from aiograpi.mixins.highlight import HighlightMixin
from aiograpi.mixins.igtv import DownloadIGTVMixin, UploadIGTVMixin
from aiograpi.mixins.insights import InsightsMixin
from aiograpi.mixins.location import LocationMixin
from aiograpi.mixins.media import MediaMixin
from aiograpi.mixins.multiple_accounts import MultipleAccountsMixin
from aiograpi.mixins.note import NoteMixin
from aiograpi.mixins.notification import NotificationMixin
from aiograpi.mixins.password import PasswordMixin
from aiograpi.mixins.photo import DownloadPhotoMixin, UploadPhotoMixin
from aiograpi.mixins.private import PrivateRequestMixin
from aiograpi.mixins.public import (
    ProfilePublicMixin,
    PublicRequestMixin,
    TopSearchesPublicMixin,
)
from aiograpi.mixins.share import ShareMixin
from aiograpi.mixins.signup import SignUpMixin
from aiograpi.mixins.story import StoryMixin
from aiograpi.mixins.timeline import ReelsMixin
from aiograpi.mixins.totp import TOTPMixin
from aiograpi.mixins.track import TrackMixin
from aiograpi.mixins.user import UserMixin
from aiograpi.mixins.video import DownloadVideoMixin, UploadVideoMixin


class Client(
    MultipleAccountsMixin,
    NoteMixin,
    GraphQLRequestMixin,
    PublicRequestMixin,
    ChallengeResolveMixin,
    PrivateRequestMixin,
    TopSearchesPublicMixin,
    ProfilePublicMixin,
    LoginMixin,
    ShareMixin,
    TrackMixin,
    FbSearchMixin,
    HighlightMixin,
    DownloadPhotoMixin,
    UploadPhotoMixin,
    DownloadVideoMixin,
    UploadVideoMixin,
    DownloadAlbumMixin,
    NotificationMixin,
    UploadAlbumMixin,
    DownloadIGTVMixin,
    UploadIGTVMixin,
    MediaMixin,
    UserMixin,
    InsightsMixin,
    CollectionMixin,
    AccountMixin,
    DirectMixin,
    LocationMixin,
    HashtagMixin,
    CommentMixin,
    StoryMixin,
    PasswordMixin,
    DownloadClipMixin,
    UploadClipMixin,
    ReelsMixin,
    BloksMixin,
    TOTPMixin,
    SignUpMixin,
):
    proxy = None
    logger = logging.getLogger("aiograpi")

    def __init__(
        self, settings: dict = {}, proxy: str = None, delay_range: list = None, **kwargs
    ):
        super().__init__(**kwargs)
        self.settings = settings
        self.delay_range = delay_range
        self.set_proxy(proxy)
        self.init()

    def set_proxy(self, dsn: str):
        if not dsn:
            self.public.proxies = self.private.proxies = {}
            return False
        assert isinstance(
            dsn, str
        ), f'Proxy must been string (URL), but now "{dsn}" ({type(dsn)})'
        self.proxy = dsn
        proxy_href = "{scheme}{href}".format(
            scheme="http://" if not urlparse(self.proxy).scheme else "",
            href=self.proxy,
        )
        self.public.proxies = self.private.proxies = self.graphql.proxies = {
            "all://": proxy_href
        }
        return True



import json
from pathlib import Path
from typing import List
from urllib.parse import urlparse

from aiograpi import config
from aiograpi.exceptions import (
    ClientNotFoundError,
    PreLoginRequired,
    PrivateError,
    StoryNotFound,
    UserNotFound,
)
from aiograpi.extractors import extract_story_gql, extract_story_v1, extract_user_short
from aiograpi.types import Story, UserShort


class StoryMixin:
    def story_pk_from_url(self, url: str) -> str:
        """
        Get Story (media) PK from URL

        Parameters
        ----------
        url: str
            URL of the story

        Returns
        -------
        str
            Media PK

        Examples
        --------
        https://www.instagram.com/stories/dhbastards/2581281926631793076/
          -> 2581281926631793076
        """
        path = urlparse(url).path
        parts = [p for p in path.split("/") if p and p.isdigit()]
        return str(parts[0])

    # def story_info_gql(self, story_pk: str):
    #     # GQL havent video_url :-(
    #     return self.media_info_gql(self, str(story_pk))

    async def story_info_v1(self, story_pk: str) -> Story:
        """
        Get Story by pk or id

        Parameters
        ----------
        story_pk: str
            Unique identifier of the story

        Returns
        -------
        Story
            An object of Story type
        """
        story_id = await self.media_id(story_pk)
        story_pk, user_id = story_id.split("_")
        stories = await self.user_stories_v1(user_id)
        for story in stories:
            if story_pk == story.pk:
                return story
        raise StoryNotFound(story_pk=story_pk, **self.last_json)

    async def story_info(self, story_pk: str) -> Story:
        """
        Get Story by pk or id

        Parameters
        ----------
        story_pk: str
            Unique identifier of the story

        Returns
        -------
        Story
            An object of Story type
        """
        return await self.story_info_v1(story_pk)

    async def story_delete(self, story_pk: str) -> bool:
        """
        Delete story

        Parameters
        ----------
        story_pk: str
            Unique identifier of the story

        Returns
        -------
        bool
            A boolean value
        """
        if not self.user_id:
            raise PreLoginRequired
        media_id = await self.media_id(story_pk)
        return await self.media_delete(media_id)

    async def users_stories_gql(
        self, user_ids: List[str], amount: int = 0
    ) -> List[UserShort]:
        """
        Get a user's stories (Public API)

        Parameters
        ----------
        user_ids: List[str]
            List of users
        amount: int
            Max amount of stories

        Returns
        -------
        List[UserShort]
            A list of objects of UserShort for each user_id
        """
        self.inject_sessionid_to_public()

        def _userid_chunks():
            if user_ids is None:
                raise Exception("Undefined user_ids")
            user_ids_per_query = 50
            for i in range(0, len(user_ids), user_ids_per_query):
                end = i + user_ids_per_query
                yield user_ids[i:end]

        stories_un = {}
        for userid_chunk in _userid_chunks():
            res = await self.public_graphql_request(
                query_hash="303a4ae99711322310f25250d988f3b7",
                variables={"reel_ids": userid_chunk, "precomposed_overlay": False},
            )
            stories_un.update(res)
        users = []
        for media in stories_un["reels_media"]:
            user = extract_user_short(media["owner"])
            items = media["items"]
            if amount:
                items = items[:amount]
            user.stories = [extract_story_gql(m) for m in items]
            users.append(user)
        return users

    async def user_stories_gql(
        self, user_id: str, amount: int = None
    ) -> List[UserShort]:
        """
        Get a user's stories (Public API)

        Parameters
        ----------
        user_id: str
        amount: int, optional
            Maximum number of story to return, default is all

        Returns
        -------
        List[UserShort]
            A list of objects of UserShort for each user_id
        """
        user = await self.users_stories_gql([user_id], amount=amount)[0]
        stories = user.stories
        if amount:
            stories = stories[:amount]
        return stories

    async def user_stories_v1(self, user_id: str, amount: int = None) -> List[Story]:
        """
        Get a user's stories (Private API)

        Parameters
        ----------
        user_id: str
        amount: int, optional
            Maximum number of story to return, default is all

        Returns
        -------
        List[Story]
            A list of objects of Story
        """
        params = {
            "supported_capabilities_new": json.dumps(config.SUPPORTED_CAPABILITIES)
        }
        user_id = str(user_id)
        resp = await self.private_request(f"feed/user/{user_id}/story/", params=params)
        reel = resp.get("reel") or {}
        stories = []
        for item in reel.get("items", []):
            stories.append(extract_story_v1(item))
        if amount:
            stories = stories[: int(amount)]
        return stories

    async def user_stories(self, user_id: str, amount: int = None) -> List[Story]:
        """
        Get a user's stories

        Parameters
        ----------
        user_id: str
        amount: int, optional
            Maximum number of story to return, default is all

        Returns
        -------
        List[Story]
            A list of objects of STory
        """
        try:
            return await self.user_stories_gql(user_id, amount)
        except ClientNotFoundError as e:
            raise UserNotFound(e, user_id=user_id, **self.last_json)
        except IndexError:
            return []
        except PrivateError as e:
            raise e
        except Exception:
            return await self.user_stories_v1(user_id, amount)

    async def story_seen(self, story_pks: List[str], skipped_story_pks: List[str] = []):
        """
        Mark a story as seen

        Parameters
        ----------
        story_pks: List[str]
        skipped_story_pks: List[str]

        Returns
        -------
        bool
            A boolean value
        """
        return await self.media_seen(
            [(await self.media_id(mid)) for mid in story_pks],
            [(await self.media_id(mid)) for mid in skipped_story_pks],
        )

    async def story_download(
        self, story_pk: str, filename: str = "", folder: Path = ""
    ) -> Path:
        """
        Download story media by media_type

        Parameters
        ----------
        story_pk: str

        Returns
        -------
        Path
            Path for the file downloaded
        """
        story_pk = str(story_pk)
        story = await self.story_info(story_pk)
        url = story.thumbnail_url if story.media_type == 1 else story.video_url
        return await self.story_download_by_url(url, filename, folder)

    async def story_download_by_url(
        self, url: str, filename: str = "", folder: Path = ""
    ) -> Path:
        """
        Download story media using URL

        Parameters
        ----------
        url: str
            URL for a media
        filename: str, optional
            Filename for the media
        folder: Path, optional
            Directory in which you want to download the album, default
            is "" and will download the files to working directory

        Returns
        -------
        Path
            Path for the file downloaded
        """
        fname = urlparse(url).path.rsplit("/", 1)[1].strip()
        if not fname:
            raise Exception("The URL must contain the path to the file (mp4 or jpg)")
        filename = "%s.%s" % (filename, fname.rsplit(".", 1)[1]) if filename else fname
        path = Path(folder) / filename
        response = await self.public.get(url)
        response.raise_for_status()
        with open(path, "wb") as f:
            f.write(response.read())
        return path.resolve()

    async def story_viewers(self, story_pk: str, amount: int = 0) -> List[UserShort]:
        """
        List of story viewers (Private API)

        Parameters
        ----------
        story_pk: str
        amount: int, optional
            Maximum number of story viewers

        Returns
        -------
        List[UserShort]
            A list of objects of UserShort
        """
        users = []
        next_max_id = None
        story_pk = await self.media_pk(story_pk)
        params = {
            "supported_capabilities_new": json.dumps(config.SUPPORTED_CAPABILITIES)
        }
        while True:
            try:
                if next_max_id:
                    params["max_id"] = next_max_id
                result = await self.private_request(
                    f"media/{story_pk}/list_reel_media_viewer/", params=params
                )
                for item in result["users"]:
                    users.append(extract_user_short(item))
                if amount and len(users) >= amount:
                    break
                next_max_id = result.get("next_max_id") or result.get("max_id")
                if not next_max_id:
                    break
            except Exception as e:
                self.logger.exception(e)
                break
        if amount:
            users = users[: int(amount)]
        return users

    async def story_like(self, story_id: str, revert: bool = False) -> bool:
        """
        Like a story

        Parameters
        ----------
        story_id: str
            Unique identifier of a Story
        revert: bool, optional
            If liked, whether or not to unlike. Default is False

        Returns
        -------
        bool
            A boolean value
        """
        if not self.user_id:
            raise PreLoginRequired
        media_id = await self.media_id(story_id)
        data = {
            "media_id": media_id,
            "_uid": str(self.user_id),
            "source_of_like": "button",
            "tray_session_id": self.tray_session_id,
            "viewer_session_id": self.client_session_id,
            "container_module": "reel_feed_timeline",
        }
        name = "unsend" if revert else "send"
        result = await self.private_request(
            f"story_interactions/{name}_story_like", self.with_action_data(data)
        )
        return result["status"] == "ok"

    async def story_unlike(self, story_id: str) -> bool:
        """
        Unlike a story

        Parameters
        ----------
        story_id: str
            Unique identifier of a Story

        Returns
        -------
        bool
            A boolean value
        """
        return await self.story_like(story_id, revert=True)

    async def sticker_tray(self) -> dict:
        """
        Getting a sticker tray from Instagram

        Returns
        -------
        dict
            Sticker Tray
        """
        data = {
            "_uid": self.user_id,
            "type": "static_stickers",
            "_uuid": self.uuid,
        }
        result = await self.private_request(
            "creatives/sticker_tray/",
            data=data,
            with_signature=True,
        )
        if result["status"] != "ok":
            raise Exception("Sticker not ok")
        return result



import asyncio
import base64
from datetime import datetime
from uuid import uuid4

from aiograpi.exceptions import InvalidNonce
from aiograpi.extractors import extract_user_short
from aiograpi.mixins.challenge import ChallengeChoice
from aiograpi.types import UserShort
from aiograpi.utils import generate_jazoest


class SignUpMixin:
    waterfall_id = str(uuid4())
    adid = str(uuid4())
    wait_seconds = 5

    async def signup(
        self,
        username: str,
        password: str,
        email: str = "",
        phone_number: str = "",
        full_name: str = "",
        year: int = None,
        month: int = None,
        day: int = None,
    ) -> UserShort:
        if not (email or phone_number):
            raise Exception("Use email or phone_number")
        check = await self.check_username(username)
        if not check.get("available"):
            raise Exception(f"Username is't available ({check})")
        kwargs = {
            "username": username,
            "password": password,
            "full_name": full_name,
            "year": year,
            "month": month,
            "day": day,
        }
        if email:
            kwargs["email"] = email
            check = await self.check_email(email)
            if not check.get("valid"):
                raise Exception(f"Email not valid ({check})")
            if not check.get("available"):
                raise Exception(f"Email not available ({check})")
            config = await self.get_signup_config()
            sent = await self.send_verify_email(email)
            if not sent.get("email_sent"):
                raise Exception("Email not sent ({sent})")
            # send code confirmation
            code = ""
            for attempt in range(1, 11):
                code = await self.challenge_code_handler(
                    username, ChallengeChoice.EMAIL
                )
                if code:
                    break
                await asyncio.sleep(self.wait_seconds * attempt)
            print(
                f'Enter code "{code}" for {username} '
                f"({attempt} attempts, by {self.wait_seconds} seconds)"
            )
            kwargs["email_code"] = self.check_confirmation_code(email, code).get(
                "signup_code"
            )
        if phone_number:
            kwargs["phone_number"] = phone_number
            config = await self.get_signup_config()
            check = await self.check_phone_number(phone_number)
            if check.get("status") != "ok":
                raise Exception(f"Phone number not valid ({check})")
            sms = await self.send_signup_sms_code(phone_number)
            if check.get("status") != "ok":
                raise Exception(f"Error when verify phone number ({sms})")
            if "verification_code" in sms:
                # when you have multiple accounts
                code = sms["verification_code"]
            else:
                for attempt in range(1, 11):
                    code = await self.challenge_code_handler(
                        username, ChallengeChoice.SMS
                    )
                    if code:
                        break
                    await asyncio.sleep(self.wait_seconds * attempt)
            print(
                f'Enter code "{code}" for {username} '
                f"({attempt} attempts, by {self.wait_seconds} seconds)"
            )
            kwargs["phone_code"] = code
        retries = 0
        if "tos_version" in config:
            kwargs["tos_version"] = config["tos_version"]
        while retries < 3:
            data = await self.accounts_create(**kwargs)
            if data.get("error_type") == "invalid_nonce":
                raise InvalidNonce(data["errors"]["nonce"][0])
            if data.get("message") != "challenge_required":
                break  # SUCCESS EXIT
            if self.challenge_flow(data["challenge"]):
                kwargs.update({"suggestedUsername": "", "sn_result": "MLA"})
            retries += 1
        self.authorization_data = self.parse_authorization(
            self.last_response.headers.get("ig-set-authorization")
        )
        try:
            return extract_user_short(data["created_user"])
        except Exception as e:
            print(f"ERROR: {e}", data)

    async def check_username(self, username):
        return await self.private_request(
            "users/check_username/", data={"username": username, "_uuid": self.uuid}
        )

    async def get_signup_config(self) -> dict:
        return await self.private_request(
            "consent/get_signup_config/",
            params={"guid": self.uuid, "main_account_selected": False},
        )

    async def check_email(self, email) -> dict:
        """Check available (free, not registred) email"""
        return await self.private_request(
            "users/check_email/",
            {
                "android_device_id": self.android_device_id,
                "login_nonce_map": "{}",
                "login_nonces": "[]",
                "email": email,
                "qe_id": str(uuid4()),
                "waterfall_id": self.waterfall_id,
            },
        )

    async def send_verify_email(self, email) -> dict:
        """Send request to receive code to email"""
        return await self.private_request(
            "accounts/send_verify_email/",
            {
                "phone_id": self.phone_id,
                "device_id": self.android_device_id,
                "email": email,
                "waterfall_id": self.waterfall_id,
                "auto_confirm_only": "false",
            },
        )

    async def check_confirmation_code(self, email, code) -> dict:
        """Enter code from email"""
        return await self.private_request(
            "accounts/check_confirmation_code/",
            {
                "code": code,
                "device_id": self.android_device_id,
                "email": email,
                "waterfall_id": self.waterfall_id,
            },
        )

    async def check_age_eligibility(self, year, month, day):
        return await self.private.post(
            "consent/check_age_eligibility/",
            data={"_csrftoken": self.token, "day": day, "year": year, "month": month},
        ).json()

    async def accounts_create(
        self,
        username: str,
        password: str,
        email: str = "",
        email_code: str = "",  # signup_code
        phone_number: str = "",
        phone_code: str = "",  # verification_code
        full_name: str = "",
        year: int = None,
        month: int = None,
        day: int = None,
        **kwargs,
    ) -> dict:
        timestamp = datetime.now().strftime("%s")
        self.username = username
        self.password = password
        data = {
            "jazoest": generate_jazoest(self.phone_id),
            "suggestedUsername": "",
            "do_not_auto_login_if_credentials_match": "true",  # C#
            "phone_id": self.phone_id,
            "enc_password": await self.password_encrypt(password),
            "username": username,
            "first_name": str(full_name),
            "adid": self.adid,
            "guid": self.uuid,
            "year": str(year),
            "month": str(month),
            "day": str(day),
            "device_id": self.android_device_id,
            "_uuid": self.uuid,
            "waterfall_id": self.waterfall_id,
            "one_tap_opt_in": "true",  # C#
            "_csrftoken": self.token,  # C#
            # "sn_result": "GOOGLE_PLAY_UNAVAILABLE:SERVICE_INVALID",  #C#
            **kwargs,
        }
        if self.user_id:
            data["logged_in_user_authorization_token"] = self.authorization
            data["logged_in_user_id"] = str(self.user_id)
        if email and not phone_number:  # EMAIL
            endpoint = "accounts/create/"
            # $"{emailOrPhoneNumber}|{DateTimeHelper.ToUnixTime(DateTime.UtcNow)}|{Encoding.UTF8.GetString(b)}";
            nonce = f'{email}|{timestamp}|\xb9F"\x8c\xa2I\xaaz|\xf6xz\x86\x92\x91Y\xa5\xaa#f*o%\x7f'
            sn_nonce = base64.encodebytes(nonce.encode()).decode().strip()
            data = dict(
                data,
                **{
                    "email": email,
                    "is_secondary_account_creation": "true",  # C#
                    "force_sign_up_code": email_code,
                    "sn_nonce": sn_nonce,
                    "qs_stamp": "",  # C#
                },
            )
        else:  # PHONE
            endpoint = "accounts/create_validated/"
            # $"{emailOrPhoneNumber}|{DateTimeHelper.ToUnixTime(DateTime.UtcNow)}|{Encoding.UTF8.GetString(b)}";
            nonce = f'{phone_number}|{timestamp}|\xb9F"\x8c\xa2I\xaaz|\xf6xz\x86\x92\x91Y\xa5\xaa#f*o%\x7f'
            sn_nonce = base64.encodebytes(nonce.encode()).decode().strip()
            data = dict(
                data,
                **{
                    "phone_number": phone_number,
                    "is_secondary_account_creation": "true"
                    if data.get("logged_in_user_id")
                    else "false",
                    "verification_code": phone_code,
                    "force_sign_up_code": "",
                    "has_sms_consent": "true",
                    # "sn_nonce": sn_nonce
                },
            )
        return await self.private_request(endpoint, data)

    async def challenge_flow(self, data):
        data = await self.challenge_api(data)
        while True:
            if data.get("message") == "challenge_required":
                data = await self.challenge_captcha(data["challenge"])
                continue
            elif data.get("challengeType") == "SubmitPhoneNumberForm":
                data = await self.challenge_submit_phone_number(data)
                continue
            elif data.get("challengeType") == "VerifySMSCodeFormForSMSCaptcha":
                data = await self.challenge_verify_sms_captcha(data)
                continue

    async def challenge_api(self, data):
        resp = await self.private.get(
            f"https://i.instagram.com/api/v1{data['api_path']}",
            params={
                "guid": self.uuid,
                "device_id": self.android_device_id,
                "challenge_context": data["challenge_context"],
            },
        )
        return resp.json()

    async def challenge_captcha(self, data):
        g_recaptcha_response = await self.captcha_resolve()
        resp = await self.private.post(
            f"https://i.instagram.com{data['api_path']}",
            data={"g-recaptcha-response": g_recaptcha_response},
        )
        return resp.json()

    async def challenge_submit_phone_number(self, data, phone_number):
        api_path = data.get("navigation", {}).get("forward")
        resp = await self.private.post(
            f"https://i.instagram.com{api_path}",
            data={
                "phone_number": phone_number,
                "challenge_context": data["challenge_context"],
            },
        )
        return resp.json()

    async def challenge_verify_sms_captcha(self, data, security_code):
        api_path = data.get("navigation", {}).get("forward")
        resp = await self.private.post(
            f"https://i.instagram.com{api_path}",
            data={
                "security_code": security_code,
                "challenge_context": data["challenge_context"],
            },
        )
        return resp.json()

    async def check_phone_number(self, phone_number: str):
        resp = await self.private_request(
            "accounts/check_phone_number/",
            data={
                "phone_id": self.phone_id,
                "login_nonce_map": "{}",
                "phone_number": phone_number.replace(" ", "+"),
                "guid": self.uuid,
                "device_id": self.android_device_id,
                "prefill_shown": "False",
            },
        )
        return resp

    async def send_signup_sms_code(self, phone_number: str):
        resp = await self.private_request(
            "accounts/send_signup_sms_code/",
            data={
                "phone_id": self.phone_id,
                "phone_number": phone_number.replace(" ", "+"),
                "guid": self.uuid,
                "device_id": self.android_device_id,
                "android_build_type": "release",
                "waterfall_id": self.waterfall_id,
            },
        )
        return resp



import asyncio
import json
import logging
import time

import orjson

from aiograpi import reqwests
from aiograpi.exceptions import (
    AboutUsError,
    AccountSuspended,
    ChallengeRequired,
    ClientBadRequestError,
    ClientConnectionError,
    ClientError,
    ClientForbiddenError,
    ClientGraphqlError,
    ClientJSONDecodeError,
    ClientLoginRequired,
    ClientNotFoundError,
    ClientThrottledError,
    ClientUnauthorizedError,
    TermsAccept,
    TermsUnblock,
)
from aiograpi.utils import random_delay


class PublicRequestMixin:
    public_requests_count = 0
    PUBLIC_API_URL = "https://www.instagram.com/"
    GRAPHQL_PUBLIC_API_URL = "https://www.instagram.com/graphql/query/"
    last_public_response = None
    last_public_json = {}
    request_logger = logging.getLogger("public_request")
    last_response_ts = 0
    # Need some timeout for avoid sockets and memory leaks.
    # Remember - igerl timeout 45sec
    max_read_timeout = 46

    def __init__(self, *args, **kwargs):
        self.public = reqwests.Session()
        self.public.verify = False  # fix SSLError/HTTPSConnectionPool
        self.public.headers.update(
            {
                "Connection": "Keep-Alive",
                "Accept": "*/*",
                "Accept-Encoding": "gzip,deflate",
                "Accept-Language": "en-US",
                "User-Agent": (
                    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) "
                    "AppleWebKit/605.1.15 (KHTML, like Gecko) "
                    "Version/11.1.2 Safari/605.1.15"
                ),
            }
        )
        super().__init__(*args, **kwargs)

    async def public_request(
        self,
        url,
        data=None,
        params=None,
        headers=None,
        return_json=False,
        retries_count=1,
        retries_timeout=2,
    ):
        kwargs = dict(
            data=data,
            params=params,
            headers=headers,
            return_json=return_json,
        )
        if retries_count > 10:
            raise Exception("Retries count is too high")
        if retries_timeout > 600:
            raise Exception("Retries timeout is too high")
        for iteration in range(retries_count):
            try:
                if self.delay_range:
                    await random_delay(delay_range=self.delay_range)
                return await self._send_public_request(url, **kwargs)
            except (
                ClientLoginRequired,
                ClientNotFoundError,
                ClientBadRequestError,
            ) as e:
                raise e  # Stop retries
            # except JSONDecodeError as e:
            #     raise ClientJSONDecodeError(e, respones=self.last_public_response)
            except ClientError as e:
                msg = str(e)
                if all(
                    (
                        isinstance(e, ClientConnectionError),
                        "SOCKSHTTPSConnectionPool" in msg,
                        "Max retries exceeded with url" in msg,
                        "Failed to establish a new connection" in msg,
                    )
                ):
                    raise e
                if retries_count > iteration + 1:
                    await asyncio.sleep(retries_timeout)
                else:
                    raise e
                continue

    async def _send_public_request(
        self, url, data=None, params=None, headers=None, return_json=False
    ):
        self.last_public_response = None
        self.public_requests_count += 1
        if headers:
            self.public.headers.update(headers)
        if self.last_response_ts and (time.time() - self.last_response_ts) < 1.0:
            await asyncio.sleep(1.0)
        try:
            if data is not None:
                response = await self.public.post(url, data=data, params=params)
            else:
                response = await self.public.get(url, params=params)
            self.request_logger.debug(
                "public_request %s: %s", response.status_code, response.url
            )
            self.request_logger.info(
                "[%s] [%s] %s %s",
                self.public.proxies.get("https"),
                response.status_code,
                "POST" if data else "GET",
                response.url,
            )
            self.last_public_response = response
            response.raise_for_status()
            if return_json:
                self.last_public_json = response.json()
                return self.last_public_json
            return response.text

        except orjson.JSONDecodeError as e:
            url = str(response.url)
            if "/login/" in url:
                raise ClientLoginRequired(e, response=response)
            elif "/challenge/" in url:
                raise ChallengeRequired(e, response=response)
            elif "/suspended/" in url:
                raise AccountSuspended(e, response=response)
            elif "/terms/unblock" in url:
                raise TermsUnblock(e, response=response)
            elif "/terms/accept" in url:
                raise TermsAccept(e, response=response)
            elif "/about-us" in url:
                raise AboutUsError(e, response=response)

            self.request_logger.error(
                "Status %s: JSONDecodeError in public_request (url=%s) >>> %s",
                response.status_code,
                url,
                response.text,
            )
            raise ClientJSONDecodeError(
                "JSONDecodeError {0!s} while opening {1!s}".format(e, url),
                response=response,
            )
        except reqwests.HTTPError as e:
            match getattr(self.last_public_response, "status_code", None):
                case 401:
                    exc = ClientUnauthorizedError
                case 403:
                    exc = ClientForbiddenError
                case 400:
                    exc = ClientBadRequestError
                case 429:
                    exc = ClientThrottledError
                case 404:
                    exc = ClientNotFoundError
                case _:
                    exc = ClientError
            raise exc(e, response=self.last_public_response)
        except (reqwests.ConnectError, reqwests.ReadError) as e:
            raise ClientConnectionError("{} {}".format(e.__class__.__name__, str(e)))
        finally:
            self.last_response_ts = time.time()

    async def public_a1_request(
        self, endpoint, data=None, params=None, headers=None, full=False
    ):
        url = self.PUBLIC_API_URL + endpoint.lstrip("/")
        params = params or {}
        params |= {"__a": 1, "__d": "dis"}
        response = await self.public_request(
            url, data=data, params=params, headers=headers, return_json=True
        )
        if full:
            return response
        return response.get("graphql") or response

    async def public_graphql_request(
        self,
        variables,
        query_hash=None,
        query_id=None,
        data=None,
        params=None,
        headers=None,
    ):
        if not (query_id or query_hash):
            raise Exception("Must provide valid one of: query_id, query_hash")
        default_params = {"variables": json.dumps(variables, separators=(",", ":"))}
        if query_id:
            # 17851374694183129
            default_params["query_id"] = query_id
        if query_hash:
            # 7dd9a7e2160524fd85f50317462cff9f
            default_params["query_hash"] = query_hash
        if params:
            params.update(default_params)
        else:
            params = default_params

        try:
            body_json = await self.public_request(
                self.GRAPHQL_PUBLIC_API_URL,
                data=data,
                params=params,
                headers=headers,
                return_json=True,
            )

            if body_json.get("status", None) != "ok":
                raise ClientGraphqlError(
                    "Unexpected status '{}' in response. Message: '{}'".format(
                        body_json.get("status", None), body_json.get("message", None)
                    ),
                    response=body_json,
                )

            return body_json["data"]

        except ClientBadRequestError as e:
            message = None
            try:
                body_json = e.response.json()
                message = body_json.get("message", None)
            except orjson.JSONDecodeError:
                pass
            raise ClientGraphqlError(
                "Error: '{}'. Message: '{}'".format(e, message), response=e.response
            )


class TopSearchesPublicMixin:
    async def top_search(self, query):
        """Anonymous IG search request"""
        url = "https://www.instagram.com/web/search/topsearch/"
        params = {
            "context": "blended",
            "query": query,
            "rank_token": 0.7763938004511706,
            "include_reel": "true",
        }
        return await self.public_request(url, params=params, return_json=True)


class ProfilePublicMixin:
    async def location_feed(self, location_id, count=16, end_cursor=None):
        if count > 50:
            raise ValueError("Count cannot be greater than 50")
        variables = {
            "id": location_id,
            "first": int(count),
        }
        if end_cursor:
            variables["after"] = end_cursor
        data = await self.public_graphql_request(
            variables, query_hash="1b84447a4d8b6d6d0426fefb34514485"
        )
        return data["location"]

    async def profile_related_info(self, profile_id):
        variables = {
            "user_id": profile_id,
            "include_chaining": True,
            "include_reel": True,
            "include_suggested_users": True,
            "include_logged_out_extras": True,
            "include_highlight_reels": True,
            "include_related_profiles": True,
        }
        data = await self.public_graphql_request(
            variables, query_hash="e74d51c10ecc0fe6250a295b9bb9db74"
        )
        return data["user"]



import random
from typing import List, Optional, Tuple

from aiograpi.exceptions import (  # CommentsDisabled,
    ClientError,
    ClientLoginRequired,
    ClientNotFoundError,
    CommentNotFound,
    MediaNotFound,
    PreLoginRequired,
    PrivateError,
)
from aiograpi.extractors import extract_comment
from aiograpi.mixins.graphql import GQL_STUFF
from aiograpi.types import Comment
from aiograpi.utils import dumps, generate_jazoest


class CommentMixin:
    """
    Helpers for managing comments on a Media
    """

    async def media_comments_threaded_gql_chunk(
        self, media_pk: str, comment_pk: str, end_cursor: str = ""
    ) -> Tuple[List[dict], str]:
        """
        Get threaded comments on a media

        Parameters
        ----------
        comment_pk: str
            Unique identifier of a Comment
        end_cursor: str
            Cursor

        Returns
        -------
        Tuple[List[dict], str]
            A list of objects of Comment
        """
        doc_id = "7171917939589632"
        comments = []

        # variables = {"comment_id": comment_pk, "first": 50}
        # if end_cursor:
        #     variables["after"] = end_cursor
        # data = await self.public_graphql_request(
        #     variables, query_hash="1ee91c32fc020d44158a3192eda98247"
        # )
        # comment = data.get("comment")
        # if not comment:
        #     raise CommentNotFound(**data)
        # edge_threaded_comments = comment.get("edge_threaded_comments", {})
        # for edge in edge_threaded_comments.get("edges", []):
        #     comments.append(edge["node"])
        # page_info = edge_threaded_comments["page_info"]
        # end_cursor = page_info["end_cursor"] if page_info["has_next_page"] else None

        media_pk = str(await self.media_pk(media_pk))
        variables = {
            "after": end_cursor or None,
            "before": None,
            "first": 50,
            "last": None,
            "media_id": media_pk,
            "parent_comment_id": str(comment_pk),
            "is_chronological": None,
        }
        data = {
            "variables": dumps(variables),
            "doc_id": doc_id,
            "fb_dtsg": await self.fb_dtsg,
            # OPTIONAL (may have random values):
            "jazoest": generate_jazoest(self.phone_id),
            **GQL_STUFF,
        }
        resp = await self.graphql_request(data=data)
        if data := resp["data"]:
            key = None
            for key in data.keys():
                if "comments" in key:
                    break
            item = data[key]
            edges = item.get("edges", [])
            if not edges:
                raise CommentNotFound(**data)
            for edge in edges:
                comments.append(edge["node"])
            page_info = item.get("page_info", {})
            end_cursor = (
                page_info.get("end_cursor") if page_info.get("has_next_page") else None
            )
            return comments, end_cursor
        return [], ""

    async def media_comments_threaded_gql(
        self, media_pk: str, comment_pk: str, amount: int = 0
    ) -> List[dict]:
        """
        Get threaded comments on a media

        Parameters
        ----------
        comment_pk: str
            Unique identifier of a Comment
        amount: int, optional
            Maximum number of media to return, default is 0 - Inf

        Returns
        -------
        List[dict]
            A list of objects of Comment
        """
        end_cursor = ""
        comments = []
        while True:
            items, end_cursor = await self.media_comments_threaded_gql_chunk(
                media_pk, comment_pk, end_cursor=end_cursor
            )
            comments.extend(items)
            if not end_cursor or len(items) == 0:
                break
            if amount and len(comments) >= amount:
                break
        if amount:
            comments = comments[:amount]
        return comments

    async def media_comments_gql_chunk(
        self, media_pk: str, end_cursor: str = ""
    ) -> Tuple[List[dict], str]:
        """
        Get comments on a media

        Parameters
        ----------
        media_pk: str
            Unique identifier of a Media
        end_cursor: str
            Cursor

        Returns
        -------
        Tuple[List[dict], str]
            A list of objects of Comment
        """
        media_pk = str(await self.media_pk(media_pk))
        comments = []

        # shortcode = await self.media_code_from_pk(media_pk)
        # variables = {
        #     "shortcode": shortcode,
        #     "child_comment_count": 50,
        #     "fetch_comment_count": 50,
        #     "parent_comment_count": 50,
        #     "has_threaded_comments": True,
        # }
        # query_hash = "477b65a610463740ccdb83135b2014db"
        # if end_cursor:
        #     variables = {"shortcode": shortcode, "first": 50, "after": end_cursor}
        #     query_hash = "bc3296d1ce80a24b1b6e40b1e72903f5"
        # data = await self.public_graphql_request(variables, query_hash=query_hash)
        # shortcode_media = data.get("shortcode_media")
        # if not shortcode_media:
        #     raise MediaNotFound(media_pk=media_pk, **data)
        # if shortcode_media.get("comments_disabled"):
        #     raise CommentsDisabled(media_pk=media_pk, **data)
        # edge_media_to_parent_comment = shortcode_media.get(
        #     "edge_media_to_parent_comment", {}
        # )
        # for edge in edge_media_to_parent_comment.get("edges", []):
        #     comments.append(edge["node"])
        # page_info = edge_media_to_parent_comment.get("page_info", {})
        # end_cursor = (
        #     page_info.get("end_cursor") if page_info.get("has_next_page") else None
        # )

        doc_id = "6974885689225067"
        variables = {
            "after": end_cursor or None,
            "before": None,
            "first": 50,
            "last": None,
            "media_id": media_pk,
            "sort_order": "popular",
        }
        data = {
            "variables": dumps(variables),
            "doc_id": doc_id,
            "fb_dtsg": await self.fb_dtsg,
            # OPTIONAL (may have random values):
            "jazoest": generate_jazoest(self.phone_id),
            **GQL_STUFF,
        }
        resp = await self.graphql_request(data=data)
        if data := resp["data"]:
            key = None
            for key in data.keys():
                if "comments" in key:
                    break
            item = data[key]
            for edge in item.get("edges", []):
                comments.append(edge["node"])
            page_info = item.get("page_info", {})
            end_cursor = (
                page_info.get("end_cursor") if page_info.get("has_next_page") else None
            )
            return comments, end_cursor
        return [], ""

    async def media_comments_gql(
        self, media_pk: str, amount: int = 50, max_requests: int = 0
    ) -> List[dict]:
        """
        Get comments on a media

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media
        amount: int, optional
            Maximum number of media to return, default is 0 - Inf

        Returns
        -------
        List[dict]
            A list of objects of Comment
        """
        media_pk = await self.media_pk(media_pk)
        end_cursor = ""
        comments = []
        i = 0
        while True:
            i += 1
            if max_requests and i > max_requests:
                break
            items, end_cursor = await self.media_comments_gql_chunk(
                media_pk, end_cursor=end_cursor
            )
            comments.extend(items)
            if not end_cursor or len(items) == 0:
                break
            if amount and len(comments) >= amount:
                break
        if amount:
            comments = comments[:amount]
        return comments

    async def media_stream_comments_v1_chunk(
        self, media_id: str, min_id: str = "", max_id: str = ""
    ) -> Tuple[List[Comment], str, str]:
        """
        Get comments on a media

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media

        Returns
        -------
        List[Comment]
            A list of objects of Comment
        """
        params = {
            "can_support_threading": "true",
            "inventory_source": "explore_story",
            "analytics_module": "comments_v2_feed_timeline",
            "is_carousel_bumped_post": "false",
            "feed_position": "1",
        }
        if min_id:
            params["min_id"] = min_id
        if max_id:
            params["max_id"] = max_id
        result = await self.private_request(
            f"media/{media_id}/stream_comments/",
            params=params,
        )
        # if result.get("has_more_comments"):
        #     params = {"max_id": max_id}
        # else:  # has_more_headload_comments
        #     params = {"min_id": min_id}
        comments = []
        rows = result.get("stream_rows") or [result]
        for row in rows:
            for comment in row["comments"]:
                comments.append(extract_comment(comment))
        min_id = result.get("next_min_id") or result.get("min_id", "")
        max_id = result.get("next_max_id") or result.get("max_id", "")
        return comments, min_id, max_id

    async def media_comments_v1_chunk(
        self, media_id: str, min_id: str = "", max_id: str = ""
    ) -> Tuple[List[Comment], str, str]:
        """
        Get comments on a media

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media

        Returns
        -------
        List[Comment]
            A list of objects of Comment
        """
        params = {"can_support_threading": "true", "permalink_enabled": "false"}
        if min_id:
            params["min_id"] = min_id
        if max_id:
            params["max_id"] = max_id
        result = await self.private_request(
            f"media/{media_id}/comments/", params=params
        )
        # if result.get("has_more_comments"):
        #     params = {"max_id": max_id}
        # else:  # has_more_headload_comments
        #     params = {"min_id": min_id}
        comments = []
        for comment in result.get("comments", []):
            comments.append(extract_comment(comment))
        min_id = result.get("next_min_id") or result.get("min_id", "")
        max_id = result.get("next_max_id") or result.get("max_id", "")
        return comments, min_id, max_id

    async def media_comments_v1(self, media_id: str, amount: int = 20) -> List[Comment]:
        """
        Get comments on a media

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media
        amount: int, optional
            Maximum number of media to return, default is 0 - Inf

        Returns
        -------
        List[Comment]
            A list of objects of Comment
        """
        comments = []
        min_id, max_id = None, None
        try:
            while True:
                items, min_id, max_id = await self.media_comments_v1_chunk(
                    media_id, min_id, max_id
                )
                comments.extend(items)
                if len(items) == 0:
                    break
                if amount and len(comments) > amount:
                    break
        except ClientNotFoundError as e:
            raise MediaNotFound(e, media_id=media_id, **self.last_json)
        except ClientError as e:
            if "Media not found" in str(e):
                raise MediaNotFound(e, media_id=media_id, **self.last_json)
            raise e
        if amount:
            comments = comments[:amount]
        return comments

    async def media_comments(self, media_id: str, amount: int = 20) -> List[Comment]:
        try:
            try:
                comments = await self.media_comments_gql(media_id, amount)
            except ClientLoginRequired as e:
                if not self.inject_sessionid_to_public():
                    raise e
                comments = await self.media_comments_gql(media_id, amount)  # retry
        except PrivateError as e:
            raise e
        except Exception as e:
            if not isinstance(e, ClientError):
                self.logger.exception(e)  # Register unknown error
            # Restricted Video: This video is not available in your country.
            # Or private account
            comments = await self.media_comments_v1(media_id, amount)
        return comments

    async def media_comment(
        self, media_id: str, text: str, replied_to_comment_id: Optional[int] = None
    ) -> Comment:
        """
        Post a comment on a media

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media
        text: str
            String to be posted on the media

        Returns
        -------
        Comment
            An object of Comment type
        """
        if not self.user_id:
            raise PreLoginRequired
        data = {
            "delivery_class": "organic",
            "feed_position": "0",
            "container_module": "self_comments_v2_feed_contextual_self_profile",
            "user_breadcrumb": self.gen_user_breadcrumb(len(text)),
            "idempotence_token": self.generate_uuid(),
            "comment_text": text,
        }
        if replied_to_comment_id:
            data["replied_to_comment_id"] = int(replied_to_comment_id)
        result = await self.private_request(
            f"media/{media_id}/comment/",
            self.with_action_data(data),
        )
        return extract_comment(result["comment"])

    async def comment_like(self, comment_pk: int, revert: bool = False) -> bool:
        """
        Like a comment on a media

        Parameters
        ----------
        comment_pk: int
            Unique identifier of a Comment
        revert: bool, optional
            If liked, whether or not to unlike. Default is False

        Returns
        -------
        bool
            A boolean value
        """
        if not self.user_id:
            raise PreLoginRequired
        comment_pk = int(comment_pk)
        data = {
            "is_carousel_bumped_post": "false",
            "container_module": "feed_contextual_self_profile",
            "feed_position": str(random.randint(0, 6)),
        }
        name = "unlike" if revert else "like"
        result = await self.private_request(
            f"media/{comment_pk}/comment_{name}/", self.with_action_data(data)
        )
        return result["status"] == "ok"

    async def comment_unlike(self, comment_pk: int) -> bool:
        """
        Unlike a comment on a media

        Parameters
        ----------
        comment_pk: int
            Unique identifier of a Comment

        Returns
        -------
        bool
            A boolean value
        """
        return await self.comment_like(comment_pk, revert=True)

    async def comment_bulk_delete(self, media_id: str, comment_pks: List[int]) -> bool:
        """
        Delete a comment on a media

        Parameters
        ----------
        media_id: str
            Unique identifier of a Media
        comment_pks: List[int]
            List of unique identifier of a Comment

        Returns
        -------
        bool
            A boolean value
        """
        media_id = await self.media_id(media_id)
        data = {
            "comment_ids_to_delete": ",".join([str(pk) for pk in comment_pks]),
            "container_module": "self_comments_v2_newsfeed_you",
        }
        result = await self.private_request(
            f"media/{media_id}/comment/bulk_delete/", self.with_action_data(data)
        )
        return result["status"] == "ok"

    async def comment_likers_gql_chunk(
        self, comment_pk: str, end_cursor: str = ""
    ) -> Tuple[List[dict], str]:
        """
        Get likers of comment

        Parameters
        ----------
        comment_pk: str
            Unique identifier of a Media
        end_cursor: str
            Cursor

        Returns
        -------
        Tuple[List[dict], str]
            A list of objects of Users
        """
        comment_pk = str(comment_pk)
        self.inject_sessionid_to_public()
        likers = []
        variables = {
            "comment_id": str(comment_pk),
            "first": 50,
        }
        query_hash = "5f0b1f6281e72053cbc07909c8d154ae"
        if end_cursor:
            variables["after"] = end_cursor
        data = await self.public_graphql_request(variables, query_hash=query_hash)
        comment = data.get("comment") or {}
        edge_liked_by = comment.get("edge_liked_by") or {}
        for edge in edge_liked_by.get("edges") or []:
            likers.append(edge["node"])
        end_cursor = ""
        if "page_info" in edge_liked_by:
            page_info = edge_liked_by["page_info"]
            end_cursor = page_info["end_cursor"] if page_info["has_next_page"] else None
        return likers, end_cursor

    async def comment_likers_gql(self, comment_pk: str, amount: int = 0) -> List[dict]:
        """
        Get likers of comment

        Parameters
        ----------
        comment_pk: str
            Unique identifier of a Media
        end_cursor: str
            Cursor

        Returns
        -------
        List[dict]
            A list of objects of Users
        """
        comment_pk = str(comment_pk)
        end_cursor = ""
        likers = []
        while True:
            items, end_cursor = await self.comment_likers_gql_chunk(
                comment_pk, end_cursor=end_cursor
            )
            likers.extend(items)
            if not end_cursor or len(items) == 0:
                break
            if amount and len(likers) >= amount:
                break
        if amount:
            likers = likers[:amount]
        return likers



import json
import random
import time
from pathlib import Path
from typing import Dict, List
from urllib.parse import urlparse

from aiograpi import config
from aiograpi.exceptions import HighlightNotFound
from aiograpi.extractors import extract_highlight_v1
from aiograpi.types import Highlight
from aiograpi.utils import dumps


class HighlightMixin:
    async def highlight_pk_from_url(self, url: str) -> str:
        """
        Get Highlight PK from URL

        Parameters
        ----------
        url: str
            URL of highlight

        Returns
        -------
        str
            Highlight PK

        Examples
        --------
        https://www.instagram.com/stories/highlights/17895485201104054/ -> 17895485201104054
        """
        assert "/highlights/" in url, 'URL must contain "/highlights/"'
        path = urlparse(url).path
        parts = [p for p in path.split("/") if p and p.isdigit()]
        return str(parts[0])

    async def user_highlights_v1(
        self, user_id: int, amount: int = 0
    ) -> List[Highlight]:
        """
        Get a user's highlight

        Parameters
        ----------
        user_id: int
        amount: int, optional
            Maximum number of highlight to return, default is 0 (all highlights)

        Returns
        -------
        List[Highlight]
            A list of objects of Highlight
        """
        amount = int(amount)
        user_id = int(user_id)
        params = {
            "supported_capabilities_new": json.dumps(config.SUPPORTED_CAPABILITIES),
            "phone_id": self.phone_id,
            "battery_level": random.randint(25, 100),
            "is_charging": random.randint(0, 1),
            "will_sound_on": random.randint(0, 1),
        }
        result = await self.private_request(
            f"highlights/{user_id}/highlights_tray/", params=params
        )
        return [extract_highlight_v1(highlight) for highlight in result.get("tray", [])]

    async def user_highlights(self, user_id: int, amount: int = 0) -> List[Highlight]:
        """
        Get a user's highlights

        Parameters
        ----------
        user_id: int
        amount: int, optional
            Maximum number of highlight to return, default is 0 (all highlights)

        Returns
        -------
        List[Highlight]
            A list of objects of Highlight
        """
        return await self.user_highlights_v1(user_id, amount)

    async def highlight_info_v1(self, highlight_pk: str) -> Highlight:
        """
        Get Highlight by pk or id (by Private Mobile API)

        Parameters
        ----------
        highlight_pk: str
            Unique identifier of Highlight

        Returns
        -------
        Highlight
            An object of Highlight type
        """
        highlight_id = f"highlight:{highlight_pk}"
        data = {
            "exclude_media_ids": "[]",
            "supported_capabilities_new": json.dumps(config.SUPPORTED_CAPABILITIES),
            "source": "profile",
            "_uid": str(self.user_id),
            "_uuid": self.uuid,
            "user_ids": [highlight_id],
        }
        result = await self.private_request("feed/reels_media/", data)
        data = result["reels"]
        if str(highlight_pk) not in (k.split(":")[-1] for k in data.keys()):
            raise HighlightNotFound(highlight_pk=highlight_pk, **data)
        item = next(item for item in data.values() if isinstance(item, dict))
        return extract_highlight_v1(item)

    async def highlight_info(self, highlight_pk: str) -> Highlight:
        """
        Get Highlight by pk or id

        Parameters
        ----------
        highlight_pk: str
            Unique identifier of Highlight

        Returns
        -------
        Highlight
            An object of Highlight type
        """
        return await self.highlight_info_v1(highlight_pk)

    async def highlight_create(
        self,
        title: str,
        story_ids: List[str],
        cover_story_id: str = "",
        crop_rect: List[float] = [0.0, 0.21830457, 1.0, 0.78094524],
    ) -> Highlight:
        """
        Create highlight

        Parameters
        ----------
        title: str
            Title
        story_ids: List[str]
            List of story ids
        cover_story_id: str
            User story as cover, default is first of story_ids

        Returns
        -------
        Highlight
            An object of Highlight type
        """
        if not cover_story_id:
            cover_story_id = story_ids[0]
        data = {
            "supported_capabilities_new": json.dumps(config.SUPPORTED_CAPABILITIES),
            "source": "self_profile",
            "creation_id": str(int(time.time())),
            "_uid": str(self.user_id),
            "_uuid": self.uuid,
            "cover": dumps(
                {
                    "media_id": await self.media_id(cover_story_id),
                    "crop_rect": dumps(crop_rect),
                }
            ),
            "title": title,
            "media_ids": dumps([await self.media_id(sid) for sid in story_ids]),
        }
        result = await self.private_request("highlights/create_reel/", data=data)
        return extract_highlight_v1(result["reel"])

    async def highlight_edit(
        self,
        highlight_pk: str,
        title: str = "",
        cover: Dict = {},
        added_media_ids: List[str] = [],
        removed_media_ids: List[str] = [],
    ):
        data = {
            "supported_capabilities_new": json.dumps(config.SUPPORTED_CAPABILITIES),
            "source": "self_profile",
            "_uid": str(self.user_id),
            "_uuid": self.uuid,
            "added_media_ids": dumps(added_media_ids),
            "removed_media_ids": dumps(removed_media_ids),
        }
        if title:
            data["title"] = title
        if cover:
            data["cover"] = dumps(cover)
        result = await self.private_request(
            f"highlights/highlight:{highlight_pk}/edit_reel/", data=data
        )
        return extract_highlight_v1(result["reel"])

    async def highlight_change_title(self, highlight_pk: str, title: str) -> Highlight:
        """
        Change title for highlight

        Parameters
        ----------
        highlight_pk: str
            Unique identifier of Highlight
        title: str
            Title of Highlight

        Returns
        -------
        Highlight
        """
        return await self.highlight_edit(highlight_pk, title=title)

    async def highlight_change_cover(
        self, highlight_pk: str, cover_path: Path
    ) -> Highlight:
        """
        Change cover for highlight

        Parameters
        ----------
        highlight_pk: str
            Unique identifier of Highlight
        cover_path: Path
            Path to photo

        Returns
        -------
        Highlight
        """
        upload_id, width, height = await self.photo_rupload(Path(cover_path))
        cover = {"upload_id": str(upload_id), "crop_rect": "[0.0,0.0,1.0,1.0]"}
        return await self.highlight_edit(highlight_pk, cover=cover)

    async def highlight_add_stories(
        self, highlight_pk: str, added_media_ids: List[str]
    ) -> Highlight:
        """
        Add stories to highlight

        Parameters
        ----------
        highlight_pk: str
            Unique identifier of Highlight
        removed_media_ids: List[str]
            Remove stories from highlight

        Returns
        -------
        Highlight
        """
        return await self.highlight_edit(highlight_pk, added_media_ids=added_media_ids)

    async def highlight_remove_stories(
        self, highlight_pk: str, removed_media_ids: List[str]
    ) -> Highlight:
        """
        Remove stories from highlight

        Parameters
        ----------
        highlight_pk: str
            Unique identifier of Highlight
        removed_media_ids: List[str]
            Remove stories from highlight

        Returns
        -------
        Highlight
        """
        return self.highlight_edit(highlight_pk, removed_media_ids=removed_media_ids)

    async def highlight_delete(self, highlight_pk: str) -> bool:
        """
        Delete highlight

        Parameters
        ----------
        highlight_pk: str
            Unique identifier of Highlight

        Returns
        -------
        bool
        """
        data = {"_uid": str(self.user_id), "_uuid": self.uuid}
        result = await self.private_request(
            f"highlights/highlight:{highlight_pk}/delete_reel/", data=data
        )
        return result.get("status") == "ok"
